import logging, os, json
from urllib.parse import urlunparse, quote
import cherrypy
from cherrypy._cpnative_server import CPHTTPServer
from cherrypy.process.plugins import Daemonizer

from lmsmusly.utils import config

_LOGGER = logging.getLogger(__name__)

class WebUI(object):

	def __init__(self, mus, mta, lms):
		_LOGGER.debug( "static dir: {}".format(config.WEBUI_STATICPATH ))
		self._mus = mus
		self._mta = mta
		self._lms = lms
	
	def start_server(self):
		appconf = {
			'/': {
				'tools.sessions.on': True,
				'tools.staticdir.root': config.WEBUI_STATICPATH
				},
			'/api': {
				'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
				'tools.response_headers.on': True,
				'tools.response_headers.headers': [('Content-Type', 'text/plain')],
				},
			'/api/similar': {
				'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
				'tools.response_headers.on': True,
				'tools.response_headers.headers': [('Content-Type', 'text/plain')],
				},
			'/static': {
				'tools.staticdir.on': True,
				'tools.staticdir.dir': config.WEBUI_STATICDIR
				}
		}
		#_LOGGER.debug( "app conf: {}".format( repr( appconf)))
		servconf = {'server.socket_port': config.WEBUI_PORT,
							'server.socket_host': config.WEBUI_HOST,
							'checker.on': False,
							'engine.autoreload.on': False,
							'tools.log_headers.on': False,
							'log.screen': False,
#							'log.access_file': os.path.join(config.WEBUI_LOGPATH, 'access.log'),
							'log.access_file': '',
							'log.error_file': os.path.join(config.WEBUI_LOGPATH, 'lmweb-error.log'),
#							'global': {
#								'environment' : 'test_suite'
#								'environment' : 'production'
#							}
						}
#		_LOGGER.debug( "server conf #1: {}".format( repr( servconf)))
		logging.getLogger("cherrypy").propagate = False
#		cherrypy.server.httpserver = CPHTTPServer(cherrypy.server)
		webapp = MainPage()
		webapp.api = WebService( self._mus, self._mta, self._lms)
		webapp.api.similar = WebServiceJsonPsimilars( self._mus, self._mta, self._lms)
		webapp.api.getstats = WebServiceJsonPstats( self._mus)
		cherrypy.tree.mount( webapp, "/", appconf)
		cherrypy.config.update( servconf)
#		_LOGGER.debug( "server conf #2: {}".format( repr( cherrypy.config)))
		cherrypy.engine.start()
		_LOGGER.info( "webserver started at {}:{} (logs: {})".format(config.WEBUI_HOST, config.WEBUI_PORT, config.WEBUI_LOGPATH ))

	def get_cherrypy(self):
		return cherrypy


class MainPage(object):
	@cherrypy.expose
	def index(self):
		return open( os.path.join(config.WEBUI_STATICPATH, config.WEBUI_STATICDIR, 'main.html'))

@cherrypy.expose
class WebService(object):
	def __init__(self, mus, mta, lms):
		self._mus = mus
		self._mta = mta
		self._lms = lms
	
	@cherrypy.tools.accept(media='text/plain')
	@cherrypy.tools.json_out()
	def GET(self, cmd, path = None, numres = 10, callback = None):
		_LOGGER.debug("WebService: cmd: {}".format(cmd))
		if cmd == "lm-exit":
			cherrypy.engine.publish("lm-exit")
		elif cmd == "lm-pause":
			cherrypy.engine.publish("lm-pause")
		elif cmd == "lm-unpause":
			cherrypy.engine.publish("lm-unpause")
		# FIXME duplicate code
		elif cmd == "lm-getstats":
			fs = []
			stats = self._mus.get_stats()
			for key, value in config.STATS_LABELS.items():
				if key in stats:
					fs.append( {'label':value, 'value': stats[ key]})
			return { 'status':'ok', 'stats_raw': stats, 'stats_formatted': fs }
		# FIXME duplicate code
		elif cmd == "lm-getsimilar":
			_LOGGER.debug( "WebService getsimilar path: {} numRes: {})".format( path, numres))
			if path == None:
				return { 'cmd': cmd, 'status':'error'}
			try:
				sid = self._mta.paths.index( path)
				_LOGGER.debug( "WebService getsimilar sid: {})".format( sid))
				( rt, rsims) = self._mus.get_similars( self._mta.mtracks, self._mta.mtrackids, sid, int(numres)+1)
				if rt == None:
					return { 'cmd': cmd, 'status':'error', 'errmsg': 'No results'}
				tracks = []
				for i in range( 1, len(rt)):
					url = urlunparse( ('file', '', quote( self._mta.paths[ rt[i]]), '', '', '') )
					tracks.append( { 'path': self._mta.paths[ rt[i]], 'url': url, 'sim': rsims[i], 'rank': i})
				if callback == None:
					return { 'cmd': cmd, 'status':'ok', 'tracks': tracks, "count": len( tracks)}
				else:
					return '{0}({1})'.format(callback, {'status':'ok', 'tracks': tracks})
			except ValueError:
				errmsg = "WebService: track {} not found in database".format( path)
				_LOGGER.warn( errmsg)
				return { 'cmd': cmd, 'status':'error', 'errmsg': errmsg }

		return { 'cmd': cmd, 'status':'ok' }

@cherrypy.expose
class WebServiceJsonPsimilars(object):
	def __init__(self, mus, mta, lms):
		self._mus = mus
		self._mta = mta
		self._lms = lms
	
	@cherrypy.tools.accept(media='text/plain')
	def GET(self, path, callback, numres = 10):
		_LOGGER.debug( "WebService JSONP getsimilar path: {} numRes: {})".format( path, numres))
		if path == None:
			return { 'cmd': "/api/similar", 'status':'error'}
		try:
			sid = self._mta.paths.index( path)
			_LOGGER.debug( "WebService getsimilar sid: {})".format( sid))
			( rt, rsims) = self._mus.get_similars( self._mta.mtracks, self._mta.mtrackids, sid, int(numres)+1)
			if rt == None:
				return { 'cmd': "/api/similar", 'status':'error', 'errmsg': 'No results'}
			tracks = []
			for i in range( 1, len(rt)):
				url = urlunparse( ('file', '', quote( self._mta.paths[ rt[i]]), '', '', '') )
				tracks.append( { 'path': self._mta.paths[ rt[i]], 'url': url, 'sim': rsims[i], 'rank': i})
			return '{0}({1})'.format(callback, json.dumps( {'status':'ok', 'tracks': tracks}))
		except ValueError:
			errmsg = "Not in database: {}".format( path)
			_LOGGER.warn( errmsg)
			return '{0}({1})'.format(callback, json.dumps( {'cmd': "/api/similar", 'status':'error', 'errmsg': errmsg}))

@cherrypy.expose
class WebServiceJsonPstats(object):
	def __init__(self, mus):
		self._mus = mus

	@cherrypy.tools.accept(media='text/plain')
	def GET(self, callback):
		_LOGGER.debug( "WebService JSONP getstats")
		fs = []
		stats = self._mus.get_stats()
		for key, value in config.STATS_LABELS.items():
			if key in stats:
				fs.append( {'label':value, 'value': stats[ key]})
		return '{0}({1})'.format(callback, json.dumps( { 'status':'ok', 'stats_raw': stats, 'stats_formatted': fs }))
